# Ignore PEP8 protected-access to client class | pylint: disable=W0212
"""
This module contains all tests for the CaseExporter() class
"""
import shutil
from pathlib import Path
import pytest
from vlinder.case_importer import CaseImporter
from vlinder.case_exporter import CaseExporter


@pytest.fixture(name="export_beerwiser")
def fixture_export_beerwiser():
    """
    This fixture initialises an export class for a Beerwiser case
    """
    # get a dataframe dictionary from beerwiser -- this class is tested in test_case_importer
    case = CaseImporter(Path.cwd() / "src/vlinder/data", "beerwiser", "csv")
    input_dict, _ = case.import_case()
    # initialize exporter
    yield CaseExporter(Path.cwd(), "beerwiser", input_dict)
    # remove stuff after testing is done
    shutil.rmtree(Path.cwd() / "test_folder")


def test_create_output_folder(export_beerwiser):
    """
    This function tests _create_output_folder to create an output folder
    :param export_beerwiser: a CaseExporter() fixture of Beerwiser
    """
    export_beerwiser._create_output_folder("test_folder")
    expected_folder_location = Path.cwd() / "test_folder"
    assert expected_folder_location.exists()


def test_store_as_excel_template(export_beerwiser):
    """
    This function tests _store_as_excel_template to have stored an xlsx file. Note that the format of this file
    is NOT checked with this test, just the existence
    :param export_beerwiser: a CaseExporter() fixture of Beerwiser
    """
    export_beerwiser._create_output_folder("test_folder")
    export_beerwiser._store_as_excel_template()
    expected_file_location = Path.cwd() / "test_folder" / "beerwiser.xlsx"
    assert expected_file_location.exists()


def test_input_to_dataframe(export_beerwiser):
    """
    This function tests input_to_dataframe for Beerwiser and checks whether the created
    dataframe_dict follows the required structure defined in src/vlinder/data/template.xlsx
    :param export_beerwiser: a CaseExporter() fixture of Beerwiser
    """
    # Make folder for test
    export_beerwiser._create_output_folder("test_folder")
    # Create dataframe_dict using input_to_dataframe
    dataframe_dict = export_beerwiser.input_to_dataframe()
    # Check that keys in the dataframe_dict are correct
    assert list(dataframe_dict.keys()) == [
        "configurations",
        "generic_text_elements",
        "case_text_elements",
        "key_outputs",
        "decision_makers_options",
        "scenarios",
        "fixed_inputs",
        "dependencies",
        "theme_weights",
        "key_output_weights",
        "scenario_weights",
    ]
    # Check correctness of columns within each dataframe_dict['configurations']
    assert list(dataframe_dict["configurations"].columns) == ["configuration", "value"]
    # Check correctness of columns within each dataframe_dict['generic_text_elements']
    assert list(dataframe_dict["generic_text_elements"].columns) == ["generic_text_element", "value"]
    # Check correctness of columns within each dataframe_dict['case_text_elements']
    assert list(dataframe_dict["case_text_elements"].columns) == ["case_text_element", "value"]
    # Check correctness of columns within each dataframe_dict['key_outputs']
    assert list(dataframe_dict["key_outputs"].columns) == [
        "key_output",
        "theme",
        "monetary",
        "smaller_the_better",
        "linear",
        "automatic",
        "start",
        "end",
    ]
    # Check correctness of columns within each dataframe_dict['decision_makers_options']
    assert list(dataframe_dict["decision_makers_options"].columns) == [
        "internal_variable_input",
        "decision_makers_option",
        "value",
    ]
    # Check correctness of columns within each dataframe_dict['scenarios']
    assert list(dataframe_dict["scenarios"].columns) == ["external_variable_input", "scenario", "value"]
    # Check correctness of columns within each dataframe_dict['fixed_inputs']
    assert list(dataframe_dict["fixed_inputs"].columns) == ["fixed_input", "value"]
    # Check correctness of columns within each dataframe_dict['dependencies']
    assert list(dataframe_dict["dependencies"].columns) == ["destination", "argument_1", "argument_2", "operator"]
    # Check correctness of columns within each dataframe_dict['theme_weights']
    assert list(dataframe_dict["theme_weights"].columns) == ["theme", "weight"]
    # Check correctness of columns within each dataframe_dict['key_output_weights']
    assert list(dataframe_dict["key_output_weights"].columns) == ["key_output", "weight"]
    # Check correctness of columns within each dataframe_dict['scenario_weights']
    assert list(dataframe_dict["scenario_weights"].columns) == ["scenario", "weight"]


@pytest.mark.parametrize("requested_format", ["csv", "json", "xlsx"])
def test_create_template_for_requested_format(export_beerwiser, requested_format):
    """
    This function tests create_template_for_requested_format to create a file in the requested format. Note that only
    the existence of this is tested, NOT the content.
    :param export_beerwiser: a CaseExporter() fixture of Beerwiser
    :param requested_format:
    """
    export_beerwiser._create_output_folder("test_folder")
    export_beerwiser.create_template_for_requested_format(requested_format)
    if requested_format == "xlsx":
        expected_file_location = Path.cwd() / requested_format / "beerwiser.xlsx"
    else:
        expected_file_location = Path.cwd() / requested_format / f"configurations.{requested_format}"
    assert expected_file_location.exists()
    shutil.rmtree(Path.cwd() / requested_format)



# Ignore PEP8 protected-access to client class | pylint: disable=W0212
"""
This module contains tests for the unique logic in trbs.py
Most function are calls to other classes. The testing for that can be found in the respective test-file for that class.
"""
import pytest

from vlinder.trbs import TheResponsibleBusinessSimulator, CaseError


@pytest.fixture(name="case_beerwiser")
def fixture_case_beerwiser():
    """
    This fixture initialises a CaseImporter for the beerwiser case.
    :return: an CaseImporter class for the beerwiser case.
    """
    return TheResponsibleBusinessSimulator("Beerwiser")


def test_status_after_build(case_beerwiser):
    """
    Test to check whether building a case yields the correct status
    """
    # check if set correctly
    case_beerwiser.build()
    status_expected = {0: "build"}
    assert case_beerwiser.status == status_expected

    # and reset correctly after building again
    case_beerwiser.status[999] = "some high status code"
    case_beerwiser.build()
    assert case_beerwiser.status == status_expected


def test_evaluate_before_build(case_beerwiser):
    """
    Test to check whether evaluating a case before building raises the correct error
    """
    with pytest.raises(CaseError) as case_error:
        case_beerwiser.evaluate()
    expected_error = "Case Error: first build a case with .build()"
    assert str(case_error.value) == expected_error


def test_status_after_evaluate(case_beerwiser):
    """
    Test to check whether evaluating after building a case yields the correct status
    """
    # check if set correctly
    case_beerwiser.build()
    case_beerwiser.evaluate()
    status_expected = {0: "build", 1: "evaluate"}
    assert case_beerwiser.status == status_expected

    # and reset correctly after building again
    case_beerwiser.status[999] = "some high status code"
    case_beerwiser.evaluate()
    assert case_beerwiser.status == status_expected


def test_appreciate_too_early(case_beerwiser):
    """
    Test to check whether the correct error is raised when appreciating before build or evaluate
    """
    with pytest.raises(CaseError) as build_error:
        case_beerwiser.appreciate()

    case_beerwiser.build()
    with pytest.raises(CaseError) as evaluate_error:
        case_beerwiser.appreciate()

    assert str(build_error.value) == "Case Error: first build a case with .build()"
    assert str(evaluate_error.value) == "Case Error: first evaluate a case with .evaluate()"


def test_status_after_appreciate(case_beerwiser):
    """
    Test to check whether evaluating after building a case yields the correct status
    """
    case_beerwiser.build()
    case_beerwiser.evaluate()
    case_beerwiser.appreciate()
    status_expected = {0: "build", 1: "evaluate", 2: "appreciate"}
    assert case_beerwiser.status == status_expected


def test_visualize_before_build(case_beerwiser):
    """
    Test to check whether visualizing before building a case yields the correct error
    """
    with pytest.raises(CaseError) as case_error:
        case_beerwiser.visualize("SHOULD_NOT_MATTER", "CAN BE ANYTHING")
    expected_error = "Case Error: first build a case with .build()"
    assert str(case_error.value) == expected_error


def test_transform_before_build(case_beerwiser):
    """
    Test to check whether transforming before building a case yields the correct error
    """
    with pytest.raises(CaseError) as case_error:
        case_beerwiser.transform("SHOULD NOT MATTER")
    expected_error = "Case Error: first build a case with .build()"
    assert str(case_error.value) == expected_error


def test_modify_before_build(case_beerwiser):
    """
    Test to check whether modifying a case before building yields the correct error
    """
    with pytest.raises(CaseError) as case_error:
        case_beerwiser.modify("A", "B", "C")
    expected_error = "Case Error: first build a case with .build()"
    assert str(case_error.value) == expected_error


@pytest.mark.parametrize(
    "setup_methods, expected_error",
    [
        ([], "Case Error: first build a case with .build()"),
        (["build"], "Case Error: first evaluate a case with .evaluate()"),
        (["build", "evaluate"], "Case Error: first appreciate a case with .appreciate()"),
    ],
)
def test_make_report_or_optimize(case_beerwiser, setup_methods, expected_error):
    """
    Test to check whether asking for a report or optimisation too soon will result in the appropriate error
    """
    # Call each setup method in sequence on the case_beerwiser
    for method in setup_methods:
        getattr(case_beerwiser, method)()

    with pytest.raises(CaseError) as case_error:
        case_beerwiser.make_report("A")

    assert str(case_error.value) == expected_error

    with pytest.raises(CaseError) as case_error:
        case_beerwiser.optimize("ScenA")

    assert str(case_error.value) == expected_error



"""
This module contains the CaseExporter() class. This class deals with transforming a case or exporting the outputs.
"""

import os
from pathlib import Path
import pandas as pd


# pylint: disable=too-few-public-methods
class CaseExporter:
    """
    This class deals with the transformation into a different format and export of output of an RBS case.
    """

    def __init__(self, output_path, name, input_dict):
        self.output_path = Path(output_path)
        self.folder_name = ""
        self.name = name
        self.input_dict = input_dict
        self.dataframe_dict = self.input_to_dataframe()
        self.transformers = {
            "json": lambda table: self.dataframe_dict[table].to_json(
                self.output_path / self.folder_name / f"{table}.json", orient="table", indent=4
            ),
            "xlsx": lambda table, writer: self.dataframe_dict[table].to_excel(writer, sheet_name=table, index=False),
            "csv": lambda table: self.dataframe_dict[table].to_csv(
                self.output_path / self.folder_name / f"{table}.csv", index=False, sep=";"
            ),
        }

    def _create_output_folder(self, folder_name: str) -> None:
        """
        This function creates a new folder to store the data
        :param folder_name: name of the output folder
        """
        try:
            os.makedirs(self.output_path / folder_name)
        except FileExistsError:
            print(f"Folder '{folder_name}' already exists. ")
        self.folder_name = folder_name

    def _store_as_excel_template(self) -> None:
        """
        This function stores the data in the dataframe_dict as an Excel-file. Each table will get its own sheet.
        """
        with pd.ExcelWriter(self.output_path / self.folder_name / f"{self.name}.xlsx") as writer:
            for table, _ in self.dataframe_dict.items():
                self.transformers["xlsx"](table, writer)

    def create_template_for_requested_format(self, requested_format: str) -> None:
        """
        This function stores the data in dataframe_dict into a new files of requested_format
        :param requested_format: format of the requested file
        """
        self._create_output_folder(requested_format)
        if requested_format == "xlsx":
            self._store_as_excel_template()
        else:
            for table, _ in self.dataframe_dict.items():
                self.transformers[requested_format](table)

    def input_to_dataframe(self):
        """
        This function converts the input_dict to a dataframe_dict format by
        calling the _make_table functions
        """
        # Initialize empty dict for output
        dataframe_dict = {}
        # Add theme_weights and scenario_weights tables
        dataframe_dict.update(
            {
                "configurations": self._make_table_configurations(),
                "generic_text_elements": self._make_table_generic_text_elements(),
                "case_text_elements": self._make_table_case_text_elements(),
                "key_outputs": self._make_table_key_outputs(),
                "decision_makers_options": self._make_table_dmo(),
                "scenarios": self._make_table_scenarios(),
                "fixed_inputs": self._make_table_fixed_inputs(),
                "dependencies": self._make_table_dependencies(),
                "theme_weights": self._make_table_theme_weights(),
                "key_output_weights": self._make_table_key_output_weights(),
                "scenario_weights": self._make_table_scenario_weights(),
            }
        )
        return dataframe_dict

    def _make_table_configurations(self):
        return pd.DataFrame(
            {"configuration": self.input_dict["configurations"], "value": self.input_dict["configuration_value"]}
        )

    def _make_table_generic_text_elements(self):
        return pd.DataFrame(
            {
                "generic_text_element": self.input_dict["generic_text_elements"],
                "value": self.input_dict["generic_text_element_value"],
            }
        )

    def _make_table_case_text_elements(self):
        return pd.DataFrame(
            {
                "case_text_element": self.input_dict["case_text_elements"],
                "value": self.input_dict["case_text_element_value"],
            }
        )

    def _make_table_key_outputs(self):
        return pd.DataFrame(
            {
                "key_output": self.input_dict["key_outputs"],
                "theme": self.input_dict["key_output_theme"],
                "monetary": self.input_dict["key_output_monetary"],
                "smaller_the_better": self.input_dict["key_output_smaller_the_better"],
                "linear": self.input_dict["key_output_linear"],
                "automatic": self.input_dict["key_output_automatic"],
                "start": self.input_dict["key_output_start"],
                "end": self.input_dict["key_output_end"],
            }
        )

    def _make_table_dmo(self):
        df_wide = pd.DataFrame(
            data=self.input_dict["decision_makers_option_value"],
            index=self.input_dict["decision_makers_options"],
            columns=self.input_dict["internal_variable_inputs"],
        )
        df_long = df_wide.reset_index().melt(id_vars="index", var_name="Column_Title", value_name="Value")
        df_long.columns = ["decision_makers_option", "internal_variable_input", "value"]
        df_temp = df_long.loc[:, ["internal_variable_input", "decision_makers_option", "value"]]
        return df_temp

    def _make_table_scenarios(self):
        df_wide = pd.DataFrame(
            data=self.input_dict["scenario_value"],
            index=self.input_dict["scenarios"],
            columns=self.input_dict["external_variable_inputs"],
        )
        df_long = df_wide.reset_index().melt(id_vars="index", var_name="Column_Title", value_name="Value")
        df_long.columns = ["scenario", "external_variable_input", "value"]
        df_temp = df_long.loc[:, ["external_variable_input", "scenario", "value"]]
        return df_temp

    def _make_table_fixed_inputs(self):
        return pd.DataFrame(
            {"fixed_input": self.input_dict["fixed_inputs"], "value": self.input_dict["fixed_input_value"]}
        )

    def _make_table_dependencies(self):
        return pd.DataFrame(
            {
                "destination": self.input_dict["destination"],
                "argument_1": self.input_dict["argument_1"],
                "argument_2": self.input_dict["argument_2"],
                "operator": self.input_dict["operator"],
            }
        )

    def _make_table_theme_weights(self):
        return pd.DataFrame({"theme": self.input_dict["themes"], "weight": self.input_dict["theme_weight"]})

    def _make_table_key_output_weights(self):
        return pd.DataFrame(
            {
                "key_output": self.input_dict["key_outputs"],
                "weight": self.input_dict["key_output_weight"],
            }
        )

    def _make_table_scenario_weights(self):
        return pd.DataFrame({"scenario": self.input_dict["scenarios"], "weight": self.input_dict["scenario_weight"]})
